#ifndef XFILE_H
#define XFILE_H
#include <stdio.h>
#include "kwdset.h"

/// aliases for keyword handling (used in XFILE structure)
enum kwd_handling {ignore=0, sequent_mode = 1, line_mode=2, sect_mode_seq = 3, sect_mode_full = 4};


struct xfsection;
struct XFILE;



/**
  Structure of file description. This file type is used for xfscanf function and 
  it has extended functionality comparing with traditional FILE type - 
   - actual line and column number, 
   - file name string, 
   - error and warning management,
   - support for keywords.

  Created by Tomas Koudelka, koudelka@cml.fsv.cvut.cz,  12.2006
*/
struct XFILE
{
  FILE *file;     ///< pointer to standard file structure
  long line;      ///< actual line
  long col;       ///< actual column
  long lnfpos;    ///< actual position of actual line beginning in file
  char *fname;    ///< name of opened file
  char *lnstr;    ///< line buffer where the actual file line is stored


  int warning;    ///< flag which handles warning (0=no warning/1=print warning/2=treat warning as an error)
  long maxlnover; ///< flag with the number of last oversized line (=0 if no oversized line detected)

  kwd_handling kwdmode; ///< keyword handling mode  
  long ignorecase;      ///< flag for case sensitive(=0)/case insensitive(=1) keyword searching 

  // following data members are necessary only in case that kwdmode=sect_mode
  long index_created; ///< flag for created index of particular sections
  long id_sec;        ///< index of actual section
  xfsection *asect;   ///< pointer to actual section
  long num_sec;       ///< number of detected sections
  xfsection *sect;    ///< array with descriptions of detected sections  

  private:
   long maxlnsize; ///< max line size
  public:
   XFILE();
  ~XFILE();
   long give_maxlnsize();           ///< returns size of line buffer
   void set_maxlnsize(long lnsize); ///< sets size of line buffer
};



/**
  Structure contains data about XFILE sections. It is used 
  in case that kwdmode=sect_mode

  Created by Tomas Koudelka, koudelka@cml.fsv.cvut.cz,  12.2006
*/
struct xfsection
{
  long beg_secpos;    ///< position of section beginning in xfile (section begins at the first character after the keyword for the section begining)
  long end_secpos;    ///< position of section end in xfile  (section ends at the last character befor the keyword for the section end)
  long beg_secln;     ///< line number of section beginning
  long end_secln;     ///< line number of section end
  long beg_seccol;    ///< column number of section begining (immediately after keyword of section beginning)
  long end_seccol;    ///< column number of section end (the last position before keyword of section end)
  enumstr name;       ///< section name and enum alias
};
#endif
